'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

let _stringifySqlAST = (() => {
  var _ref2 = _asyncToGenerator(function* (parent, node, prefix, context, selections, joins, wheres) {
    switch (node.type) {
      case 'table':
        if (node.where) {
          const whereCondition = yield node.where(`${quote(node.as)}`, node.args || {}, context);
          if (whereCondition) {
            wheres.push(`${whereCondition}`);
          }
        }

        if (node.sqlJoin) {
          const joinCondition = node.sqlJoin(`${quote(parent.as)}`, `${quote(node.as)}`, node.args || {});

          joins.push(`LEFT JOIN ${node.name} AS ${quote(node.as)} ON ${joinCondition}`);
        } else if (node.joinTable) {
          if (!node.sqlJoins) throw new Error('Must set "sqlJoins" for a join table.');
          const joinCondition1 = node.sqlJoins[0](`${quote(parent.as)}`, `${quote(node.joinTableAs)}`, node.args || {});
          const joinCondition2 = node.sqlJoins[1](`${quote(node.joinTableAs)}`, `${quote(node.as)}`, node.args || {});

          joins.push(`LEFT JOIN ${node.joinTable} AS ${quote(node.joinTableAs)} ON ${joinCondition1}`, `LEFT JOIN ${node.name} AS ${quote(node.as)} ON ${joinCondition2}`);
        } else {
          joins.push(`FROM ${node.name} AS ${quote(node.as)}`);
        }

        for (let child of node.children) {
          _stringifySqlAST(node, child, parent ? prefix + node.as + '__' : prefix, context, selections, joins, wheres);
        }

        break;
      case 'column':
        selections.push(`${quote(parent.as)}.${quote(node.name)} AS ${quote(prefix + node.as)}`);
        break;
      case 'columnDeps':
        for (let name in node.names) {
          selections.push(`${quote(parent.as)}.${quote(name)} AS ${quote(prefix + node.names[name])}`);
        }
        break;
      case 'composite':
        const keys = node.name.map(function (key) {
          return `${quote(parent.as)}.${quote(key)}`;
        });

        selections.push(`CONCAT(${keys.join(', ')}) AS ${quote(prefix + node.fieldName)}`);
        break;
      case 'expression':
        const expr = node.sqlExpr(`${quote(parent.as)}`, node.args || {}, context);
        selections.push(`${expr} AS ${quote(prefix + node.as)}`);
        break;
      case 'noop':
        return;
      default:
        throw new Error('unexpected/unknown node type reached: ' + (0, _util.inspect)(node));
    }
    return { selections, joins, wheres };
  });

  return function _stringifySqlAST(_x3, _x4, _x5, _x6, _x7, _x8, _x9) {
    return _ref2.apply(this, arguments);
  };
})();

var _util = require('../util');

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

exports.default = (() => {
  var _ref = _asyncToGenerator(function* (topNode, context) {
    (0, _util.validateSqlAST)(topNode);

    let { selections, joins, wheres } = yield _stringifySqlAST(null, topNode, '', context, [], [], []);

    selections = [...new Set(selections)];

    if (!selections.length) return '';

    let sql = 'SELECT\n  ' + selections.join(',\n  ') + '\n' + joins.join('\n');
    if (wheres.length) {
      sql += '\nWHERE ' + wheres.join(' AND ');
    }
    return sql;
  });

  function stringifySqlAST(_x, _x2) {
    return _ref.apply(this, arguments);
  }

  return stringifySqlAST;
})();

function quote(str) {
  return '`' + str + '`';
}