'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

let _stringifySqlAST = (() => {
  var _ref2 = _asyncToGenerator(function* (parent, node, prefix, context, selections, joins, wheres, orders) {
    switch (node.type) {
      case 'table':
        if (node.where && !node.paginate) {
          const whereCondition = yield node.where(`"${node.as}"`, node.args || {}, context);
          if (whereCondition) {
            wheres.push(`${whereCondition}`);
          }
        }

        if (node.sqlJoin) {
          const joinCondition = node.sqlJoin(`"${parent.as}"`, `"${node.as}"`, node.args || {});

          if (node.paginate) {
            let whereCondition = node.sqlJoin(`"${parent.as}"`, node.name, node.args || {});
            if (node.where) {
              const filterCondition = yield node.where(`${node.name}`, node.args || {}, context);
              if (filterCondition) {
                whereCondition += ' AND ' + filterCondition;
              }
            }

            if (node.sortKey) {

              const { limit, orderColumns, whereCondition: whereAddendum } = interpretForKeysetPaging(node);
              if (whereAddendum) {
                whereCondition += ' AND ' + whereAddendum;
              }
              const join = `\
LEFT JOIN LATERAL (
  SELECT * FROM ${node.name}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns)}
  LIMIT ${limit}
) AS "${node.as}" ON ${joinCondition}`;
              joins.push(join);
              orders.push({
                table: node.as,
                columns: orderColumns
              });
            } else if (node.orderBy) {

              const { limit, offset, orderColumns } = interpretForOffsetPaging(node);
              const join = `\
LEFT JOIN LATERAL (
  SELECT *, count(*) OVER () AS "$total"
  FROM ${node.name}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns)}
  LIMIT ${limit} OFFSET ${offset}
) AS "${node.as}" ON ${joinCondition}`;
              joins.push(join);
              orders.push({
                table: node.as,
                columns: orderColumns
              });
            }
          } else {
            joins.push(`LEFT JOIN ${node.name} AS "${node.as}" ON ${joinCondition}`);
          }
        } else if (node.joinTable) {
          if (!node.sqlJoins) throw new Error('Must set "sqlJoins" for a join table.');
          const joinCondition1 = node.sqlJoins[0](`"${parent.as}"`, `"${node.joinTableAs}"`, node.args || {});
          const joinCondition2 = node.sqlJoins[1](`"${node.joinTableAs}"`, `"${node.as}"`, node.args || {});

          if (node.paginate) {
            let whereCondition = node.sqlJoins[0](`"${parent.as}"`, node.joinTable, node.args || {});
            if (node.where) {
              const filterCondition = yield node.where(`${node.name}`, node.args || {}, context);
              if (filterCondition) {
                whereCondition += ' AND ' + filterCondition;
              }
            }

            if (node.sortKey) {
              const { limit, orderColumns, whereCondition: whereAddendum } = interpretForKeysetPaging(node);
              if (whereAddendum) {
                whereCondition += ' AND ' + whereAddendum;
              }
              const join = `\
LEFT JOIN LATERAL (
  SELECT * FROM ${node.joinTable}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns)}
  LIMIT ${limit}
) AS "${node.joinTableAs}" ON ${joinCondition1}`;
              joins.push(join);
              orders.push({
                table: node.joinTableAs,
                columns: orderColumns
              });
            } else if (node.orderBy) {

              const { limit, offset, orderColumns } = interpretForOffsetPaging(node);
              const join = `\
LEFT JOIN LATERAL (
  SELECT *, count(*) OVER () AS "$total"
  FROM ${node.joinTable}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns)}
  LIMIT ${limit} OFFSET ${offset}
) AS "${node.joinTableAs}" ON ${joinCondition1}`;
              joins.push(join);
              orders.push({
                table: node.joinTableAs,
                columns: orderColumns
              });
            }
          } else {
            joins.push(`LEFT JOIN ${node.joinTable} AS "${node.joinTableAs}" ON ${joinCondition1}`);
          }
          joins.push(`LEFT JOIN ${node.name} AS "${node.as}" ON ${joinCondition2}`);
        } else if (node.paginate) {
          if (node.sortKey) {
            let { limit, orderColumns, whereCondition } = interpretForKeysetPaging(node);
            whereCondition = whereCondition || 'TRUE';
            if (node.where) {
              const filterCondition = yield node.where(`${node.name}`, node.args || {}, context);
              if (filterCondition) {
                whereCondition += ' AND ' + filterCondition;
              }
            }
            const join = `\
FROM (
  SELECT * FROM ${node.name}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns)}
  LIMIT ${limit}
) AS "${node.as}"`;
            joins.push(join);
            orders.push({
              table: node.as,
              columns: orderColumns
            });
          } else if (node.orderBy) {
            const { limit, offset, orderColumns } = interpretForOffsetPaging(node);
            let whereCondition = 'TRUE';
            if (node.where) {
              const filterCondition = yield node.where(`${node.name}`, node.args || {}, context);
              if (filterCondition) {
                whereCondition = filterCondition;
              }
            }
            const join = `\
FROM (
  SELECT *, count(*) OVER () AS "$total"
  FROM ${node.name}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns)}
  LIMIT ${limit} OFFSET ${offset}
) AS "${node.as}"`;
            joins.push(join);
            orders.push({
              table: node.as,
              columns: orderColumns
            });
          }
        } else {
          joins.push(`FROM ${node.name} AS "${node.as}"`);
        }

        for (let child of node.children) {
          _stringifySqlAST(node, child, parent ? prefix + node.as + '__' : prefix, context, selections, joins, wheres, orders);
        }

        break;
      case 'column':
        let parentTable = node.fromOtherTable || parent.as;
        selections.push(`"${parentTable}"."${node.name}" AS "${prefix + node.as}"`);
        break;
      case 'columnDeps':
        for (let name in node.names) {
          selections.push(`"${parent.as}"."${name}" AS "${prefix + node.names[name]}"`);
        }
        break;
      case 'composite':
        const keys = node.name.map(function (key) {
          return `"${parent.as}"."${key}"`;
        });
        selections.push(`CONCAT(${keys.join(', ')}) AS "${prefix + node.fieldName}"`);
        break;
      case 'expression':
        const expr = node.sqlExpr(`"${parent.as}"`, node.args || {}, context);
        selections.push(`${expr} AS "${prefix + node.as}"`);
        break;
      case 'noop':
        return;
      default:
        throw new Error('unexpected/unknown node type reached: ' + (0, _util.inspect)(node));
    }
    return { selections, joins, wheres, orders };
  });

  return function _stringifySqlAST(_x3, _x4, _x5, _x6, _x7, _x8, _x9, _x10) {
    return _ref2.apply(this, arguments);
  };
})();

var _util = require('../util');

var _graphqlRelay = require('graphql-relay');

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

exports.default = (() => {
  var _ref = _asyncToGenerator(function* (topNode, context) {
    (0, _util.validateSqlAST)(topNode);

    let { selections, joins, wheres, orders } = yield _stringifySqlAST(null, topNode, '', context, [], [], [], []);

    selections = [...new Set(selections)];

    if (!selections.length) return '';

    let sql = 'SELECT\n  ' + selections.join(',\n  ') + '\n' + joins.join('\n');
    if (wheres.length) {
      sql += '\nWHERE ' + wheres.join(' AND ');
    }
    if (orders.length) {
      sql += '\nORDER BY ' + stringifyOuterOrder(orders);
    }
    return sql;
  });

  function stringifySqlAST(_x, _x2) {
    return _ref.apply(this, arguments);
  }

  return stringifySqlAST;
})();

function interpretForOffsetPaging(node) {
  if (node.args && node.args.last) {
    throw new Error('Backward pagination not supported with offsets. Consider using keyset pagination instead');
  }
  const orderColumns = {};
  if (typeof node.orderBy === 'object') {
    for (let column in node.orderBy) {
      let direction = node.orderBy[column].toUpperCase();
      if (direction !== 'ASC' && direction !== 'DESC') {
        throw new Error(direction + ' is not a valid sorting direction');
      }
      orderColumns[column] = direction;
    }
  } else if (typeof node.orderBy === 'string') {
    orderColumns[node.orderBy] = 'ASC';
  } else {
    throw new Error('"orderBy" is required for pagination');
  }
  let limit = 'ALL',
      offset = 0;
  if (node.args && node.args.first) {
    limit = parseInt(node.args.first) + 1;
    if (node.args.after) {
      offset = (0, _graphqlRelay.cursorToOffset)(node.args.after) + 1;
    }
  }
  return { limit, offset, orderColumns };
}

function interpretForKeysetPaging(node) {
  const orderColumns = {};
  let descending = node.sortKey.order.toUpperCase() === 'DESC';

  if (node.args && node.args.last) {
    descending = !descending;
  }
  for (let column of (0, _util.wrap)(node.sortKey.key)) {
    orderColumns[column] = descending ? 'DESC' : 'ASC';
  }

  let limit = 'ALL',
      whereCondition = '';
  if (node.args && node.args.first) {
    limit = parseInt(node.args.first) + 1;
    if (node.args.after) {
      const cursorObj = (0, _util.cursorToObj)(node.args.after);
      validateCursor(cursorObj, (0, _util.wrap)(node.sortKey.key));
      whereCondition = sortKeyToWhereCondition(cursorObj, descending);
    }
    if (node.args.before) {
      throw new Error('Using "before" with "first" is nonsensical.');
    }
  } else if (node.args && node.args.last) {
    limit = parseInt(node.args.last) + 1;
    if (node.args.before) {
      const cursorObj = (0, _util.cursorToObj)(node.args.before);
      validateCursor(cursorObj, (0, _util.wrap)(node.sortKey.key));
      whereCondition = sortKeyToWhereCondition(cursorObj, descending);
    }
    if (node.args.after) {
      throw new Error('Using "after" with "last" is nonsensical.');
    }
  }

  return { limit, orderColumns, whereCondition };
}

function sortKeyToWhereCondition(keyObj, descending) {
  const sortColumns = [];
  const sortValues = [];
  for (let key in keyObj) {
    sortColumns.push(`"${key}"`);
    sortValues.push((0, _util.maybeQuote)(keyObj[key]));
  }
  const operator = descending ? '<' : '>';
  return `(${sortColumns.join(', ')}) ${operator} (${sortValues.join(', ')})`;
}

function orderColumnsToString(orderColumns) {
  const conditions = [];
  for (let column in orderColumns) {
    conditions.push(`"${column}" ${orderColumns[column]}`);
  }
  return conditions.join(', ');
}

function stringifyOuterOrder(orders) {
  const conditions = [];
  for (let condition of orders) {
    for (let column in condition.columns) {
      const direction = condition.columns[column];
      conditions.push(`"${condition.table}"."${column}" ${direction}`);
    }
  }
  return conditions.join(', ');
}

function validateCursor(cursorObj, expectedKeys) {
  const actualKeys = Object.keys(cursorObj);
  const expectedKeySet = new Set(expectedKeys);
  const actualKeySet = new Set(actualKeys);
  for (let key of actualKeys) {
    if (!expectedKeySet.has(key)) {
      throw new Error(`Invalid cursor. The column "${key}" is not in the sort key.`);
    }
  }
  for (let key of expectedKeys) {
    if (!actualKeySet.has(key)) {
      throw new Error(`Invalid cursor. The column "${key}" is not in the cursor.`);
    }
  }
}